from o2dbErrors import *

class PersistentObject:
    """create new object in memory or fetch from DB
       depends on got parameters (objID)"""   
    def __init__(self, connBroker, objID=None, lockForWrite=0):        
        self.CB = connBroker
        self.SQL = connBroker.getSQLstatements()
        self.attrList = []
        self.conn = None
        self.ID = objID
        self.iLock = 0      
        if objID:
            if lockForWrite: self._readObjectLocked()
            else: self._readObject()
               
    def __del__(self):
        """unlock and return connection"""
        # print "__del__ wurde aufgerufen"
        ## unlock myself, if still locked
        if self.iLock:
            #print "unlock wird aufgerufen"
            try:  # maybe DB object was just deleted -> there is nothing to unlock!!!
                self.unlock()
            except: pass
        ## return connection if still in hand
        self._returnConnection()

    def deleteDBobject(self):
        """delete object in the DB, should be followed by PO = None (destroy memory object)"""
        self.lock()

        for idx in range(len(self.attrList) - 1, -1, -1):
            self._delAttribute(self.attrList[idx])

        myID = str(self.ID)

        c = self._getCursor()
        ## "DELETE * FROM tblObjectIDs WHERE ID=" + myID
        c.execute(self.SQL.getString("DELETE_03", myID))
        c.close()


    def _readObject(self):
        """read all attributes from DB; calls attribute-constructor with attrType and attrID"""
        myID = str(self.ID)
        c = self._getCursor()
        ## "SELECT * FROM tblObjects WHERE objID=" + myID
        c.execute(self.SQL.getString("SELECT_01", myID))
        all = c.fetchall()
        if not all:
            c.close()
            raise ErrObjIDnotFound("object not found", "no attributes in 'tblObjects' for DB-object with ID '" + myID + "'")
        else:
            for i in range(len(all)):
                #print all[i]
                attr = Attribute(self.CB, all[i][1], attrID=all[i][2], cursor=c)
                self.attrList.append(attr)       
            c.close()
        
    def _readObjectLocked(self):
        self.lock()
        self._readObject()

    def _lockDBobject(self):
        myID = str(self.ID)        
        c = self._getCursor()
        ## "SELECT ID, LOCKED FROM tblObjectIDs WHERE ID=" + myID
        c.execute(self.SQL.getString("SELECT_02", myID))
        all = c.fetchall()
        if len(all) == 0: raise ErrObject("object couldn't be locked", "object ID not found")
        elif len(all) > 1: raise ErrObject("object was not locked", "found more than one object")
        locked = all[0][1]
        if not locked:
            ## lock table: "LOCK TABLES tblObjectIDs WRITE"
            c.execute(self.SQL.getString("LOCK_01"))
            ## now the object must still be unlocked - so read it again
            ## "SELECT ID, LOCKED FROM tblObjectIDs WHERE ID=" + myID
            c.execute(self.SQL.getString("SELECT_02", myID))
            ## trust now that its the only _one_ object with this ID
            one = c.fetchone()
            ## set the lock by writing the (memory)object ID
            ## "UPDATE tblObjectIDs SET LOCKED='" + str(id(self)) + "' WHERE ID=" + myID
            c.execute(self.SQL.getString("UPDATE_01", str(id(self)), myID))
            ## now object is locked in database
            ## "UNLOCK TABLES"
            c.execute(self.SQL.getString("UNLOCK_01"))
            self.iLock = 1
            c.close()
        elif locked:
            c.close()
            raise ErrObjNotLockable("object could't be locked", "DB-object '" + myID \
                                    + "' was already locked from memory-object with id = '" \
                                    + str(locked) + "'\n - <btw> i am '" + str(id(self)) + "'")
        else:
            c.close()
            raise ErrObject("error object locking", "should never occur")
            
    def _unlockDBobject(self):
        myID = str(self.ID)
        c = self._getCursor()
        ## "SELECT ID, LOCKED FROM tblObjectIDs WHERE ID=" + myID
        c.execute(self.SQL.getString("SELECT_02", myID))
        all = c.fetchall()
        if len(all) == 0:
            c.close()
            raise ErrObject("object couldn't be unlocked", "object ID not found")
        elif len(all) > 1:
            c.close()
            raise ErrObject("object was not locked", "found more than one object")
        locked = all[0][1]                
        if locked:
            ## "UPDATE tblObjectIDs SET LOCKED='0' WHERE ID=" + myID
            c.execute(self.SQL.getString("UPDATE_02", myID))
            ## now object is unlocked in database
            self.iLock = 0
            c.close()
        elif not locked:
            self.iLock = 0     ## change the wrong value in the (memory)object
            c.close()
            raise ErrObjWasNotLocked("object unlock doesn't make sense :-)", \
                                     "DB-object '" + myID + "' is free like a bird, _but_ memory-object id = '" + \
                                     str(id(self)) + "' says 'I LOCK' - WHY ??? -> lock-value is corrected")
        else:
            c.close()
            raise ErrObject("error object unlocking", "should never occur")
        
    def _getCursor(self):
        if self.conn:
            return self.conn.cursor()
        else:
            self.conn = self.CB.getConnection(id(self))
            return self.conn.cursor()

    def _returnConnection(self):
        if self.conn:
            self.conn = None
            self.CB.objReturnsConnection(id(self))

    def _getID(self):
        """for object internal use: fetches ID from DB if not present -
        for external call use getID()"""
        if self.ID:
            return self.ID
        else:
            c = self._getCursor()
            ## "INSERT INTO tblObjectIDs VALUES ()"
            c.execute(self.SQL.getString("INSERT_01"))
            ## "SELECT last_insert_id()"
            c.execute(self.SQL.getString("SELECT_03"))
            self.ID = str(c.fetchone()[0])
            c.close()
            return self.ID

    def _delAttribute(self, attr):
        c = self._getCursor()
        ## aus 'richtiger' attribut-tabelle lschen
        table = self.CB.selectTable(attr.getType(), attr.getValue())
        ## "DELETE FROM " + table + " WHERE attrID=" + str(attr.getID())
        c.execute(self.SQL.getString("DELETE_01", table, str(attr.getID())))
        ## eintrag aus 'tblObjects' lschen
        ## "DELETE FROM tblObjects WHERE attrType='" + attr.getType() + "' AND attrID=" + str(attr.getID())
        c.execute(self.SQL.getString("DELETE_02", attr.getType(), str(attr.getID())))
        ## aus der attributsliste des objektes lschen
        self.attrList.remove(attr)
        c.close()
        
    def _save(self):        
        c = self._getCursor()
        OI = self._getID()

        ## schon hier wird der volltextindex mit den neuen attributen erstellt
        ## das speichern geht mit der attribut-speicher-schleife mit
        self._updateIndex()

        while 1:
            attrTableChange = []
            for attr in self.attrList:
                if attr.isDirty():
                    try:
                        attr.save(c, OI)
                    ## changing of attribute requires changing of table in the DB ->
                    ## first collect these attributes in another list
                    except ErrAttrTableChange, error:
                        #print error.reason
                        #print error.description
                        attrTableChange.append(attr)
                
            ## if list is not empty -> some attributes want to change the table
            ## the 'while' (with saving new attributes) is repeated
            if attrTableChange:              
                for at in attrTableChange:
                    ## add new attribute and delete old
                    newType = self.CB.selectTable(at.getType(), at.getValue())
                    self.addAttribute(newType, at.getName(), at.getValue())                    
                    ## BEFORE DELETING: change value to valueDB so that old and correct table ist found
                    at.setValue(at.getValueDB())
                    ## locking already done by save() so i call internal deletion: _delAttribute()
                    self._delAttribute(at)
            else: break

        c.close()

    def _updateIndex(self):
        values = ""
        for attr in self.attrList:
            if attr.getName() == "objectindex": self._delAttribute(attr)
            elif attr.getName() == "objectcategory" or attr.getName() == "mainattributes": pass
            else:
                if attr.getType().count("CHAR") == 1:
                    for item in attr.getValue().split():
                        values += item + " "
                elif attr.getType().count("DATE") == 1:
                    values += str(attr.getValue())
        self.addAttribute("fulltextindex", "objectindex", values)


    def lock(self):
        """here resides the semantic and error handling of object locking
           which can be done without walking to DB"""
        if not self.ID: raise ErrObject("object could't be locked", \
                              "object is not persistant in the database")

        if not self.iLock: self._lockDBobject()
        elif self.iLock: raise ErrObjLockedFromMe("object will not be locked", "i lock it already")
        else: raise ErrObject("error object locking", "should never occur")
       
    def unlock(self, doit=None):
        """here resides the semantic and error handling of object unlocking
           which can be done without walking to DB
           not critical - no table is locked"""
        if not self.ID: raise ErrObject("object could't be unlocked", \
                              "object is not persistant in the database")
        if self.iLock: self._unlockDBobject()
        elif not self.iLock:
            ## schurkenlschen, obwohl nicht von diesem MEM objekt gesperrt - ONLY for testing
            if doit=="really": self._unlockDBobject()
            else:
                raise ErrObjNotLockedFromMe("object will not be unlocked", \
                         "memory object '" + str(id(self)) + "' has no lock on DB-object '" + str(self.ID) + "'")
        else: raise ErrObject("error object locking", "should never occur")
       
    def getID(self):
        """for 'external' calls - no fetching of ID if object only in memory"""
        return self.ID

    def delAttribute(self, attrName, attrValue=None):
        """
        call with name -> search in attrList and call _delAttribute() with attribut-object
        from attrList
        actually an attribut-name can be present more than one time in the object
        (e.g. 2 x Surname) -> the first found in attrList will be deleted!!!
        call with name and value -> both will be matched to find the correct attribute
        """
        ## wenn das attribut in der liste nicht gefunden wird passiert nix - super ->
        ## sollte ohnehin gelscht werden;
        ## mit der lnge der attrList wird das trotzdem als fehler erkannt und abgefangen
        for i in range(len(self.attrList)):

            if (self.attrList[i].getName() == attrName and attrValue == None) or \
               (self.attrList[i].getName() == attrName and attrValue == self.attrList[i].getValue()):
                ## try object locking
                try:
                    self.lock()
                except ErrObjectLocking, error:
                    print error.reason
                    print error.description
                    print "error number = " + str(error.number)
           
                lenAttrListBefore = len(self.attrList)
                self._delAttribute(self.attrList[i])
                lenAttrListAfter = len(self.attrList)
                ## test with length of attrList if really an attribute was deleted
                if (lenAttrListBefore - 1) != lenAttrListAfter:
                    raise ErrAttrNotDeleted("error attribute delete", "attribute was not deleted, maybe it was not here")
                        
                ## try object unlocking
                try:
                    self.unlock()
                except ErrObjectUnlocking, error:
                    print error.reason
                    print error.description
                    print "error number = " + str(error.number)
                
                break
               
    def addAttribute(self, attrType, attrName, attrValue):        
        attr = Attribute(self.CB, attrType, attrName=attrName, attrValue=attrValue)
        self.attrList.append(attr)

    def save(self):
        """write all dirty attributes to database"""
        ## if object locks its data in DB -> do save
        if self.iLock:
            self._save()
        ## if object has no ID its new -> do save
        elif not self.ID:
            self._save()
        ## now test attrList if at least one attribute is dirty AND not new
        else:
            notOneDirty = 1
            for attr in self.attrList:
                if attr.isDirty() and attr.isNotNew():
                    notOneDirty = 0
                    break
            ## notOneDirty means maybe new attributes -> save withhout locking
            ## nobody can be affected
            if notOneDirty:
                self._save()
            else:
                ## try locking
                try:
                    self.lock()
                except ErrObjectLocking, error:
                    print error.reason
                    print error.description
                    print "error number = " + str(error.number)

                self._save()
                
                ## try unlocking
                try:
                    self.unlock()
                except ErrObjectUnlocking, error:
                    print error.reason
                    print error.description
                    print "error number = " + str(error.number)

    def getAttribute(self, attrName):
        """get the atttribute-object by its name"""
        for i in range(len(self.attrList)):
            if self.attrList[i].getName() == attrName:
                return self.attrList[i]

    def getAttributeValue(self, attrName):
        """get one attribute-value of the object by attributes name"""
        for i in range(len(self.attrList)):
            if self.attrList[i].getName() == attrName:
                return self.attrList[i].getValue()

    def getAttributes(self):
        """return all attributes of the object as a list of triples"""
        tripleList = []
        for attr in self.attrList:
            if attr.getName() != "objectindex":
                tripleList.append((attr.getType(), attr.getName(), attr.getValue()))
        return tripleList
         
    def setAttributes(self, tripleList):
        """write back all attributes to MEM object since setValue() of attribute
           cares that only changed values are save to DB"""
        # eigene liste bauen mit kontrollvariable, ob attribut schon verwendet wurde
        myAttrList = []; updated = 0
        for attr in self.attrList:
            myAttrList.append([attr, updated])

        for triple in tripleList:
            type, name, value = triple

            found = 0
            for myAttr in myAttrList:
                #prfen ob dieses attribut schon upgedated wurde
                if not myAttr[1]:
                    ## check if type and name of attribute match
                    cmpChar = myAttr[0].getType()[3:7]; cmpOther = myAttr[0].getType()[3:]
                    if  (myAttr[0].getName() == name) and ((cmpChar == type) or (cmpOther == type)):
                        #wert und updated==1(JA) setzen
                        myAttr[0].setValue(value); myAttr[1] = 1
                        found = 1; break

            #wenn nicht gefunden ein neues attribut dem objekt hinzufgen
            if not found:
                self.addAttribute(type, name, value)

                
    def getAttributeList(self):
        return self.attrList


class Attribute:
    """construction depends on given parameters -
        either with type, name and value
        OR with type, id and cursor -> fetch from DB"""
    def __init__(self, connBroker, attrType, **attrs):
        self.CB = connBroker
        self.SQL = connBroker.getSQLstatements()
        self.type = attrType
        self.ID = None
        
        for k in attrs.keys():
            #print k, ':', attrs[k]
            if k == 'attrID': self.ID = attrs[k]
            elif k == 'cursor': self.curs = attrs[k]
            elif k == 'attrName': self.name = attrs[k]
            elif k == 'attrValue': self.value = self.valueDB = attrs[k]
            else: raise ErrAttribute("construction of attribute failed", "received unknown attribute - '" + k + "'")
          
        if self.ID:
            self._readAttribute()
            self.new = 0
            self.dirty = 0
        else:
            self.new = 1
            self.dirty = 1

    def _readAttribute(self):
        """reads attribute from DB using the ID"""
        table = self.type
        ## "SELECT * FROM " + table + " WHERE attrID=" + str(self.ID)
        self.curs.execute(self.SQL.getString("SELECT_04", table, str(self.ID)))
        all = self.curs.fetchall()
        self.name = all[0][1]
        ## valueDB is needed to check if changed attribute requires saving to another table
        self.value = self.valueDB = all[0][2]

    def _updateAttribute(self, curs):
        table = self.CB.selectTable(self.type, self.value)
        ## test if attribute change requires another table for saving (e.g. CHAR10 -> CHAR25)
        if table != self.CB.selectTable(self.type, self.valueDB):
            raise ErrAttrTableChange("error attribute change", \
                  "attribute '" + self.name + "':" + str(self.value) + "' has to go to another table")
        else:
            ## "UPDATE " + table + " SET attrName='" + self.name + "', attrValue='" + self.value + "' WHERE attrID=" + str(self.ID)
            curs.execute(self.SQL.getString("UPDATE_03", table, self.name, self.value, str(self.ID)))
            self.dirty = 0

    def _saveAttributeNew(self, curs, objID, value):
        """saves the attribute in the corresponding table"""
        table = self.CB.selectTable(self.type, self.value)                    

        ## "INSERT INTO " + table + " (attrName, attrValue, objID) VALUES ('" + self.name + "', '" + str(value) + "', " + str(objID) + ")"
        if self.type == "BLOB":
            curs.execute(self.SQL.getString("INSERTBLOB_02", table, self.name, value, str(objID)))
        else:
            curs.execute(self.SQL.getString("INSERT_02", table, self.name, str(value), str(objID)))
        ## "SELECT last_insert_id()"
        curs.execute(self.SQL.getString("SELECT_03"))
        self.ID = str(curs.fetchone()[0])
        self.dirty = 0
                   
    def _saveAttributeObjEntry(self, curs, objID):
        """saves the new attribute for the corresponding object in 'tblObjects'"""
        type = self.CB.selectTable(self.type, self.value)
        ## "INSERT INTO tblObjects (objID, attrType, attrID) VALUES (" + str(objID) + ", '" + type + "', " + self.ID + ")"
        curs.execute(self.SQL.getString("INSERT_03", str(objID), type, self.ID))
        self.new = 0    

    def save(self, curs, objID):
        """saves attribute; special case 'Relation' attribute"""
        type = self.CB.selectTable(self.type, self.value)[3:]
        ## save for relations -
        ## attribute can only be new because editing does not make sense
        if type == "Relation":
            if self.name == "UNI":
                self._saveAttributeNew(curs, objID, self.value)
                self.dirty = 0
                self._saveAttributeObjEntry(curs, objID)
                self.new = 0
            elif self.name == "BI":
                self._saveAttributeNew(curs, objID, self.value)
                self.dirty = 0
                self._saveAttributeObjEntry(curs, objID)
                self.new = 0
                ## save relation backwards
                ## !!! the related object doesn't know until it refreshes !!!
                self._saveAttributeNew(curs, self.value, objID)
                self.dirty = 0
                self._saveAttributeObjEntry(curs, self.value)
                self.new = 0                
            else: raise ErrAttribute("error relation attribute", "type " + self.name + " not reckognized")
            
        ## save for non-relation attribute
        else:
            if self.new:
                self._saveAttributeNew(curs, objID, self.value)
                self.dirty = 0
                self._saveAttributeObjEntry(curs, objID)
                self.new = 0
            elif self.dirty:
                self._updateAttribute(curs)
                self.dirty = 0

    def getID(self):            
        return self.ID
    def getType(self):
        return self.type
    def isDirty(self):
        return self.dirty
    def isNotNew(self):
        return not self.new
    def getName(self):
        return self.name
    def getValue(self):
        return self.value
    def getValueDB(self):
        return self.valueDB
    def setValue(self, val):
        if self.value != val:
            self.value = val
            self.dirty = 1
